; *****************************************************************************
; Junior Computer ][ Print And Hex Monitor Version 0.8
; by Joerg Walke
;
; first implementation 28.12.2021
; updated 12.5.2022
; by Joerg Walke
;
; Assembled With A65
; ***************************************************************************** 
; To Do: (maybe in this order)
;
;  tape loader, saver
;  assembler
;  ...and many more...
; *****************************************************************************

VERMAIN   	EQU     '0'    		; print monitor main version
VERSUB    	EQU     '9'    		; print monitor sub version

RESET     	EQU  	$1C1F       	; original junior monitor reset vector + 2

; Buffers *********************************************************************

STRBUF	  	EQU   	$1400    	; input string buffer at $0600 rev. 2 or $1400 rev. 3 

RBUFF     	EQU   	$1500    	; xmodem receive buffer at $0700 rev. 2 or $1500 rev. 3

; Card Base Addresses *********************************************************

IOBASE		EQU	$14		; pointer to IO card base
IOBASEL		EQU	$14		; always $00
IOBASEH		EQU	$15		; K2 = $08, K3 = $0C, K4 = $10

; GRBASE	EQU	$16		; reserved
; GRBASEL	EQU	$16		; reserved
; GRBASEH	EQU	$17		; reserved

; FLPBASE	EQU	$18		; reserved
; FLPBASE	EQU	$18		; reserved
; FLPBASE	EQU	$19		; reserved

; ACIA And Terminal Variables *************************************************

PSTR      	EQU   	$EA      	; output string Pointer
PSTRL     	EQU   	$EA      	; lower address byte of output string pointer
PSTRH     	EQU   	$EB      	; upper address byte of output string pointer   
WBUF      	EQU   	$EC      	; ACIA character output buffer

; *****************************************************************************

; 		EQU	$ED 		; unused
; 		EQU	$EE 		; unused

; XModem Transfer Variables ***************************************************

CRCL      	EQU   	$DC      	; CRC lo byte 
CRCH      	EQU   	$DD     	; CRC hi byte 

RETRYL    	EQU   	$DE      	; retry counter lo byte
RETRYH    	EQU   	$DF      	; retry counter hi byte

BLKEND      	EQU     $E0		; block end flag
BFLAG     	EQU   	$E1      	; block flag
BLKNO     	EQU   	$E2      	; block number 
ERRCNT    	EQU   	$E3      	; error counter 10 is the limit

; *****************************************************************************

; 		EQU	$E4 		; unused
; 		EQU	$E5 		; unused

; Disassembler Variables ******************************************************

OPCODE		EQU	$E0		; current opcode
LINECNT		EQU	$E1		; number of disassembled lines
ADRMODE		EQU	$E2		; addressing mode
IBYTES		EQU	$E3		; instruction byte count

; I2C Variables ***************************************************************

I2C_DATA	EQU	$E0		; current I2C data byte

; Hex Monitor Variables *******************************************************

ASCL	  	EQU   	$E6     	; ASCII list start address low
ASCH      	EQU   	$E7     	; ASCII list start address high

STOL      	EQU   	$E8     	; Store address Low
STOH      	EQU   	$E9     	; Store address High

NUML      	EQU   	$F8     	; low number byte
NUMH	  	EQU   	$F9	 	; high number byte
ADRL      	EQU   	$FA     	; last address Low
ADRH      	EQU   	$FB     	; last address High
TEMP      	EQU   	$FC     	; temp storage
YSAV      	EQU   	$FD     	; Y register storage
PDBCNT    	EQU   	$FE     	; number of printed data bytes
MODE      	EQU   	$FF     	; current edit mode

; CPU Register Save ***********************************************************

PCL		EQU	$EF		; program counter Low	
PCH		EQU	$F0		; program counter High
PREG		EQU	$F1		; processor status register
SPUSER		EQU	$F2		; stack pointer
ACC		EQU	$F3		; accumulator
YREG		EQU	$F4		; y-register
XREG		EQU	$F5		; x-register

; Clock Variables *************************************************************

DIVCHAR		EQU	$F2

; ACIA Registers **************************************************************

DATA_REG  	EQU   	$1600    	; ACIA Data Register
STAT_REG  	EQU   	$1601    	; ACIA Status Register
COMM_REG  	EQU   	$1602    	; ACIA Command Register
CTRL_REG  	EQU   	$1603    	; ACIA Control Register

; RIOT RAM Area ***************************************************************

BAUDRATE  	EQU   	$1A00    	; current baud rate
STDSAVE		EQU	$1A02		; current standard save routine
STDLOAD		EQU	$1A04		; current standard load routine
STDBEEP		EQU	$1A06		; current standard beep routine
STDIN		EQU	$1A08		; current standard input routine
STDOUT		EQU	$1A0A		; current standard output routine
STDHOME		EQU	$1A0C		; current standard home routine
STDCLS		EQU	$1A0E		; current standard cls routine
STDNORMAL	EQU	$1A10		; current standard normal routine
STDINVERSE	EQU	$1A12		; current standard inverse routine
STDFLASH	EQU	$1A14		; current standard flash routine

; Interrupt Vectors ***********************************************************

BRKUSR		EQU	$1A78		; address of user BREAK vector
NMIVECT		EQU     $1A7A           ; address of NMI vector
IRQUSR  	EQU   	$1A7C    	; address of user IRQ vector
IRQVECT  	EQU   	$1A7E    	; address of main IRQ vector

; Timer Register **************************************************************

CNTA	  	EQU	$1AF4	 	; CLK/1T no interrupt
CNTB	  	EQU	$1AF5	 	; CLK/8T no interrupt
CNTC	  	EQU	$1AF6	 	; CLK/64T no interrupt
CNTD	  	EQU	$1AF7    	; CLK/1024T no interrupt

; Port Register ***************************************************************

PAD		EQU     $1A80		; Port A Data Register
PADD		EQU	$1A81		; Port A Data Direction Register
PBD		EQU	$1A82		; Port B Data Register
PBDD		EQU	$1A83		; Port B Data Direction Register

; IO Base Addresses ***********************************************************

K2		EQU	$0800		; base address of IO select K2
K3		EQU	$0C00		; base address of IO select K3
K4		EQU	$1000		; base address of IO select K4


; VIA 1 Register Indices ******************************************************

VIA_PORTB     	EQU  	$00  		; port B input/output register
VIA_PORTA     	EQU  	$01  		; port A input/output register 
VIA_DDRB   	EQU  	$02		; port B data direction register
VIA_DDRA   	EQU  	$03		; port A data direction register
VIA_T1CL   	EQU  	$04		; 
VIA_T1CH   	EQU  	$05
VIA_T1LL   	EQU  	$06
VIA_T1LH   	EQU  	$07
VIA_T2CL   	EQU  	$08
VIA_T2CH   	EQU  	$09
VIA_SR     	EQU  	$0A		; shift register
VIA_ACR    	EQU  	$0B		; auxilary control register
VIA_PCR    	EQU  	$0C		; periheral control register
VIA_IFR    	EQU  	$0D		; interrupt flag register
VIA_IER    	EQU  	$0E		; interrupt enable register
VIA_PANOHS 	EQU  	$0F

; VIA 2 Register Indices *****************************************************

PORTB     	EQU  	$10  		; port B input/output register
PORTA     	EQU  	$11  		; port A input/output register 
DDRB   		EQU  	$12		; port B data direction register
DDRA   		EQU  	$13		; port A data direction register
T1CL   		EQU  	$14		; 
T1CH   		EQU  	$15
T1LL   		EQU  	$16
T1LH   		EQU  	$17
T2CL   		EQU  	$18
T2CH   		EQU  	$19
SR     		EQU  	$1A		; shift register
ACR    		EQU  	$1B		; auxilary control register
PCR    		EQU  	$1C		; periheral control register
IFR    		EQU  	$1D		; interrupt flag register
IER    		EQU  	$1E		; interrupt enable register
PANOHS 		EQU  	$1F

; Pointer To Language Name ****************************************************

LANGKEY		EQU     $DFF0
LANGNAME	EQU	LANGKEY+1

; Constants *******************************************************************

PROMPT    	EQU     '*'    		; prompt character
ADIV      	EQU     '-'    		; address divider
BS        	EQU     $08    		; backspace key
CR        	EQU     $0D    		; carriage return
LF	  	EQU     $0A    		; line feed
CAN		EQU     $18		; Cancel
ESC       	EQU     $1B    		; ESC
SPC		EQU     $20		; space char

SOH       	EQU     $01    		; start of header
EOT       	EQU     $04    		; end of text 
ACK       	EQU     $06    		; acknowledged
NAK       	EQU     $15    		; not acknowledged

STDTTY		EQU	$00		; standard tty terminal id
STDPPORT	EQU	$01		; standard parallel port
STDASCII	EQU	$02		; standard ascii keyboard

I2C_RTC_ADR	EQU	$68		; I2C Address of DS1307 Real Time Clock

; TEMP ************************************************************************

PPORTLOAD	EQU	$2000		; jump location for test code
PPORTSAVE	EQU	$2003		; jump location for test code



		ORG 	$E000       	; start address of Monitor

MON_COLD_START	JMP  	MAINSTART	; jump to monitor cold start
MON_WARM_START	JMP	MONINP		; jump to monitor warm start
MON_SET_STDOUT	JMP	SETSTDOUT	; jump to set standard out routine

; **** IRQ, NMI and BREAK Service Routine *************************************

; *****************************************************************************

IRQ		STA	ACC		; save current accumulator
		PLA			; get current processor status in A
		STA	PREG		; and save it
		PHA			; and push it back to stack
		AND	#$10		; mask break flag
		BNE	USRBREAK	; if break flag set, jump to user break handler
		JMP	(IRQUSR)	; else jump to user IRQ routine
USRBREAK	JMP	(BRKUSR)	
NMI		STA	ACC
		PLA
		STA	PREG
		PHA
BREAK		JSR	TTYRESET	; always reset to tty I/O
		PLP			; get last processor status
		STX	XREG		; save x-register
		STY	YREG		; save y-register
		PLA			; get last program counter low byte
		STA	PCL		; and store it
		STA	ADRL		
		PLA			; get last program counter high byte
		STA	PCH		; and store it
		STA	ADRH
		TSX			; get current stack pointer
		STX	SPUSER		; and store it
		CLD			; set binary mode
		JSR	BEEP		; error beep
		JSR	PRSTATUS	; print user program status
		LDX     #$FF
		TXS			; initialize stack pointer
		JMP	MONRESET	; and return to monitor

; **** Print Program Status ***************************************************

; *****************************************************************************

PRSTATUS	JSR	PRADDRESS	; print current program counter
		LDX	#$04		
NXTREG		JSR	SPCOUT		; print space char
		LDA	PSSTR,X		; load register label
		JSR	COUT		; and print it
		LDA	#'='		
		JSR	COUT		; print =
		LDA	PREG,X
		JSR	HEXOUT
		DEX
		BPL	NXTREG		
		RTS

; **** Set Standard Out Routine ***********************************************

; Input: A - Standard Out Identifier

; *****************************************************************************

SETSTDOUT	PHA
		CMP	#STDPPORT
		BEQ	SETPPORT
		LDX	#$00
		LDY	#$00
		BEQ	SETSTDOUTPROC
SETPPORT	LDX	#PROC_PPORT_OUT-PROC_STD_OUT
SETSTDOUTPROC	LDA	PROC_STD_OUT,X
		STA	STDOUT,Y
		INX
		INY
		CPY	#$10
		BNE	SETSTDOUTPROC
		PLA
		RTS

; **** Set Standard In Routine ************************************************

; Input: A - Standard In Identifier

; *****************************************************************************

SETSTDIN	PHA
		CMP	#STDPPORT	; set parallel port as standard in?
		BEQ	SETPPORTIN	; no, set standard in to parallel port
		LDA	#LOW SERIALIN	; load low byte of address
		BNE	STOREINADR	; branch always
SETPPORTIN	LDA	#LOW PPORTIN	; load low byte of address
STOREINADR	STA	STDIN
		LDA	#HIGH PPORTIN	; load high byte of address
		STA	STDIN+1		; should be $E0 for all std in routines
		PLA
		RTS

; **** Write To Serial Routine ***************************************************

; Input: A - Output Byte to RS232

; *****************************************************************************

SERIALOUT	PHA			; save character
		LDA  	#$10
EMPTY?		BIT  	STAT_REG	; ACIA output register empty?
		BEQ  	EMPTY?		; no, check again.
		PLA			; restore character
		STA  	DATA_REG   	; write character to ACIA
		RTS

; **** Read From Serial Routine ***********************************************

; Output: A - Input Byte from RS232
;         C - 1 char get, 0 no char get

; *****************************************************************************

SERIALIN	CLC              	; set to no chr present
		LDA	STAT_REG	
		AND	#$08		; ACIA input register full?
		BEQ	SERIALEND	; no, just exit
		LDA	DATA_REG	; yes, read character
		SEC		 	; and set C = 1, char present
SERIALEND	RTS 

; **** Read From ASCII Keyboard Routine ***************************************

; Output: A - Input Byte from Keyboard
;         C - 1 char get, 0 no char get

; *****************************************************************************

ASCIIKBD	JSR	PPORTIN		; read parallel port
		AND	#%01111111	; and clear MSB
		RTS
		
; **** Write To Parallel Port Routine *****************************************

; Input: A - Output Byte to parallel port

; *****************************************************************************

PPORTOUT	PHA			; save character
		LDA	#$BE		; initialize handshake line I/O on port b
		STA	PBDD		; PB7 = /strobe, PB6 = busy, PB5 = r/w, PB0 = speaker off
		LDA	#$86		; set handshake lines to their initial values
		STA	PBD		; r/w = L, strobe = H, PB1,PB2 = H -> hex-kbd disabled
		LDA	#$FF		; all port A lines are outputs
		STA	PADD
		PLA			; reload character in A
		PHA
		STA	PAD		; set output data
PPORTBSY?	BIT	PBD		; bussy line is high?
		BVS	PPORTBSY?	; yes, check bussy line again
		LDA	#$06		; generate strobe pulse
		STA	PBD		; set strobe line low
		LDA	$86
		STA	PBD		; set strobe line high	
		PLA			; restore character
		RTS

; **** Read From Parallel Port Routine ****************************************

; Output: A - Input Byte from parallel port
;         C - 1 char get, 0 no char get

; *****************************************************************************

PPORTIN		LDA	#$00		; initialize port A as input
		STA	PADD		
		LDA	#$7E		; initialize port B bits for read operation
		STA	PBDD		
		LDA	#$20		; set PB5 = H (READ)
		STA	PBD
		CLC			; set to no chr present
STROBE?		BIT	PBD		; test PB7 (/STROBE)
		BPL	NOSTROBE	; STROBE = L? No, just exit with C = 0
		LDA	PAD		; yes, load keyboard ASCII code from port A
		SEC			; and set C = 1, char present
NOSTROBE	RTS

; **** Write Binary Routine ***************************************************

; Input: A - Output Byte to Standard Out

; *****************************************************************************

BOUT		JMP	(STDOUT)

; **** Read Character Routine *************************************************

; Output: A - character read

; *****************************************************************************

CIN		JSR	CGET		; call standard in. Character available?
		BCC	CIN		; no, repeat 
		RTS

; **** Get Character (no wait) Routine ****************************************

; Output: A - character read from standard in
;         C - 1 char get, 0 no char get 

; *****************************************************************************		

CGET            JMP	(STDIN)

; **** Write LF Character Routine *********************************************

; *****************************************************************************

LFOUT		LDA  	#LF        	; write a LF
					; fall through to COUT

; **** Write Character Routine ************************************************

; Input: A - character to write

; *****************************************************************************

COUT 		JSR     BOUT
                CMP  	#CR        	; character was a CR?
		BEQ  	LFOUT      	; yes, also write LF
		RTS

; **** Write CR/LF To Terminal ************************************************

; *****************************************************************************

CROUT		LDA	#CR		; write a CR
		BNE	COUT

; **** Write Single Space Char To Terminal ************************************

; *****************************************************************************

SPCOUT		LDA	#$20		; write a Space char
		BNE	COUT

; **** Read String Routine ****************************************************

; Output:  (PSTRL, PSTRH) - pointer to null terminated string data

; *****************************************************************************

STRIN           LDX  	#$02		; initialize character index
BACKSPACE       DEX
		BEQ  	STRIN		; if line empty, restart
NEXTCHR         JSR  	CIN        	; get next character from input buffer
		STA  	STRBUF,X   	; store character in string buffer
		CMP  	#CR		; is it a CR?
                BEQ  	ENDSTRIN	; yes, exit
                JSR  	COUT		; echo character
		CMP  	#BS        	; backspace key?
		BEQ  	BACKSPACE  	; yes
                INX             	; advance string index
                BNE  	NEXTCHR    	; more then 255 characters? No, read next char
                LDA  	#CR        	; yes, auto new line and stop reading
		STA  	STRBUF,X   	; store CR in string buffer
ENDSTRIN        JSR  	COUT       	; send CR
	        RTS


; **** Write String Routine ***************************************************

; Input:  (PSTRL, PSTRH) - pointer to null terminated string data

; *****************************************************************************

STROUT		LDY  	#$00       	; index y is 0

; **** Write String From Index Routine ****************************************

; Input:  (PSTRL, PSTRH) - pointer to null terminated string data
;	  Y              - start index into string data

; *****************************************************************************

WRSTR		LDA  	(PSTR),Y   	; load char at string pos y
		BEQ  	ENDSTROUT  	; exit, if NULL char
		JSR  	COUT       	; write character 
		INY             	; next index
		JMP  	WRSTR      
ENDSTROUT	RTS

; **** Print A Byte In Hexadecimal ********************************************

; Input: A - data byte to print in hex

; *****************************************************************************

HEXOUT          PHA             	; save A for lower hex digit
                LSR  	A
                LSR  	A
                LSR  	A                     
                LSR  	A		 
                JSR  	HEXDIG		; write upper hex digit 
                PLA             	; write lower hex digit
					; fall through to HEXDIG

; **** Print A Single Hexadecimal Digit ***************************************

; Input: A - data nibble in Bit 0-3 to print in hex

; *****************************************************************************

HEXDIG          AND     #$0F    	; mask lower digit
                ORA     #'0'    	; add 48
                CMP     #'9'+1  	; decimal digit?
                BCC     PRHEX   	; yes, print it
                ADC     #6      	; add offset for letter digit
PRHEX		JSR     COUT
		RTS

; *****************************************************************************
; VT100 Commands
; *****************************************************************************

; **** Identify Terminal Routine **********************************************

; *****************************************************************************

IDENTIFY        LDY  	#ESCGID-STRINGP	; load sequence index
		BNE  	PRINTESC   	; jump always

; **** Set Normal Text Routine ***************************************************

; *****************************************************************************

TTY_NORMAL      LDY  	#ESCNORM-STRINGP; load sequence index
		BNE  	PRINTESC   	; jump always

; **** Inverse Text Routine ***************************************************

; *****************************************************************************

TTY_INVERSE     LDY  	#ESCINV-STRINGP ; load sequence index
		BNE  	PRINTESC   	; jump always

; **** Blink Text Routine *****************************************************

; *****************************************************************************

TTY_FLASH       LDY  	#ESCBLNK-STRINGP; load sequence index
		BNE  	PRINTESC   	; jump always

; **** Cursor Home Routine ****************************************************

; *****************************************************************************

TTY_HOME        LDY  	#ESCHOME-STRINGP; load sequence index
		BNE  	PRINTESC   	; jump always
		
; **** Clear Line Routine *****************************************************

; *****************************************************************************

CLRLINE         LDY  	#ESCCLL-STRINGP	; load sequence index
		BNE  	PRINTESC   	; jump always

; **** Clear Screen Routine ***************************************************

; *****************************************************************************

TTY_CLRSCRN     LDY  	#ESCCLS-STRINGP	; load sequence index
					; fall through to PRINTESC

; **** VT100 ESC Sequence Loader **********************************************

; *****************************************************************************

PRINTESC        JSR	LOADSTRING 
		JSR  	WRSTR
		RTS

; *****************************************************************************
; Screen Commands
; *****************************************************************************
 
; **** Set Normal Text Routine ***************************************************

; *****************************************************************************

NORMAL		JMP	(STDNORMAL)

; **** Inverse Text Routine ***************************************************

; *****************************************************************************

INVERSE		JMP	(STDINVERSE)

; **** Blink Text Routine *****************************************************

; *****************************************************************************

FLASH		JMP	(STDFLASH)

; **** Cursor Home Routine ****************************************************

; *****************************************************************************

HOME		JMP	(STDHOME)

; **** Clear Screen Routine ***************************************************

; *****************************************************************************

CLRSCRN		JMP	(STDCLS)

; **** Print Tab Routine ******************************************************

; Input: A - number of space characters to print

; *****************************************************************************

TAB		STA  	TEMP
		LDA  	#SPC		; load SPC char
PRINTTAB	JSR  	COUT		; write SPC
		DEC   	TEMP
                BNE   	PRINTTAB   	; all spaces written? No, repeat
		RTS

; **** Read Hex Number From Input String **************************************

; Input:  Y - current input string position to read from 
; Output: (NUML, NUMH) - last 8 digits of read hex number 

; *****************************************************************************

HEXINPUT	LDX   	#$00
		STX   	NUML       	; clear input value
		STX   	NUMH
NEXTDIGIT       LDA   	STRBUF,Y   	; get next input char
		CMP   	#'0'
		BCC   	NOTHEX     	; char < '0'? Yes, no hex digit
		CMP   	#':'
		BCC   	NUMDIGIT   	; char is in '0'..'9'
		AND   	#$DF	 	; uppercase chars only
HEXDIGIT	CMP   	#'A'
		BCC   	NOTHEX     	; char < 'A'? Yes, no hex digit
		CMP   	#'G'
		BCS   	NOTHEX     	; char > 'F'? Yes, no hex digit
		SBC   	#'A'-11	 	; char 'A'..'F' to value 10..16
NUMDIGIT        ASL   	A
		ASL   	A
		ASL   	A
		ASL   	A		; digit shifted to upper nibble
		LDX   	#$04      	; load shift loop counter
SHIFT		ASL   	A		; shift msb in C
		ROL   	NUML
		ROL   	NUMH
		DEX
		BNE   	SHIFT	 	; 4 times shifted? No, repeat
		INY		 	; increment string index
		BNE   	NEXTDIGIT	; branch always
NOTHEX		RTS

; **** Delay Routine **********************************************************

; Input: A - milliseconds to wait

; ***************************************************************************** 

DELAY           STA  	CNTD
LOOPDELAY	LDA  	CNTA
		BNE  	LOOPDELAY
		RTS

; **** Check ESC Routine ******************************************************

; Output: C - 1 ESC pressed, 0 ESC not pressed
; Beep if ESC pressed 

; *****************************************************************************

CHKESC		JSR     CGET            ; key pressed?
		BCC	NOTESC		; no 		
		CMP     #ESC		; ESC pressed?
		BEQ	BEEP		; yes, exit and beep.		
		CLC			; no, clear carry flag
NOTESC		RTS

; **** System Beep Routine ****************************************************

; *****************************************************************************	

BEEP		JMP	(STDBEEP)	; call standard BEEP routine

; **** Simple Beep Routine ****************************************************

; *****************************************************************************	

DOBEEP		LDA	PBDD		; save port b data direction register
		PHA
		LDX     #$60		; repeat 60 times
		LDA     #$01
		TAY
		STA     PBDD		; turn speaker on
BEEPLOOP	STY     PBD		; set PB0 high
		LDA     #$02
		JSR     DELAY		; delay of ~2ms
		DEY
		STY	PBD		; set PB0 low
		LDA	#$02
		JSR	DELAY		; delay of ~2ms
		LDY	#$01
		DEX
		BNE	BEEPLOOP	; not finished, repeat
		PLA
		STA	PBDD		; restore port b data direction register
		SEC			
		RTS

; **** Reset Standard I/O to TTY **********************************************

TTYRESET	LDA  	#80
STDIORESET	JSR	SETSTDOUT	; set standard out to tty terminal
		JSR	SETSTDIN	; set standard in to tty terminal
		RTS

; **** Auto Terminal And Baud Rate Detection Routine **************************

; *****************************************************************************	

MAINSTART	LDX     #$FF
		TXS			; initialize stack pointer
		CLD			; set binary mode
		LDA	#LOW DOBEEP	; load low byte of address of system beep
		STA	STDBEEP
		LDA	#HIGH DOBEEP	; load high byte of address of system beep
		STA	STDBEEP+1
VTDETECT	JSR     TTYRESET        ; reset to tty input/output
		JSR  	DELAY		; wait for ~128ms after reset
		SEI			; disable Interrupts
		STA  	STAT_REG   	; reset ACIA  
		LDA  	#$00
		STA  	BAUDRATE   	; initialize baud rate variable
		LDA  	#$0B       	; set ACIA to
         	STA  	COMM_REG	; no parity, no receiver echo, RTS low, no IRQ, DTR low
		LDX  	#$19		; start with 1 stop bit, 8 data bits, 2400 bps as the current baud rate 
NEXTBAUD	STX  	CTRL_REG	; set the baud rate
		JSR  	IDENTIFY	; send identify string to terminal
		LDA  	#40
		JSR  	DELAY		; wait for ~64ms
		LDA  	STAT_REG
		AND  	#$08		; ACIA input register full?
		BEQ  	NOESC	 	; no, go on
		LDA  	DATA_REG    	; read data register from ACIA
		CMP  	#ESC	 	; is it a ESC char
		BNE  	NOESC	 	; no, go on
		STX  	BAUDRATE    	; and store it
		LDX  	#$1F	 	; detection finished 
NOESC           INX  			; try next baud rate
		CPX  	#$20
		BCC  	NEXTBAUD   	; tried all baud rates?

; **** Main Initialization Routine ********************************************

; ***************************************************************************** 

INIT    	JSR     BEEP
		JSR	DETECTLNGCARD	; detect if language card is available
		LDA  	BAUDRATE   	; load selected baud 
		BNE     SETBAUDRATE	; terminal detected?		
		LDA	#$1E		
		STA  	CTRL_REG	; no, set baud rate to standard 9600 bps	
		JMP  	JCRESET		; and jump to junior monitor

SETBAUDRATE     STA  	CTRL_REG	; set detected baudraute
	
INITVECT	LDA  	#LOW IRQ	; write address of IRQ service routine to IRQ vector pointer
		STA  	IRQVECT			
		LDA	#LOW NMI	; write address of NMI service routine to NMI vector pointer
		STA	NMIVECT
		LDA	#LOW BREAK	; write address of user service routines to user vector pointers
		STA	BRKUSR		 
		STA	IRQUSR		
		LDA  	#HIGH IRQ
		STA  	IRQVECT+1
		STA	NMIVECT+1
		STA	BRKUSR+1	 
		STA	IRQUSR+1	

; **** Main Program Loop ******************************************************

; *****************************************************************************

MAIN		JSR	CGET		; clear input buffer
		JSR  	CLRLOADSTR    	; clear screen and load pointer to string table
		LDA  	#$1F
		JSR  	TAB		; send some space chars to center title 
        	LDY  	#TITLE-STRINGP 	; load title string
		JSR  	WRSTR		; and write it
		LDA	IOBASEH		; language card available?
		BEQ	SHOWMON		; no, just start monitor
		
		LDY	#IOCARD-STRINGP ; load detect message
		JSR	WRSTR		; and write it
		LDA	IOBASEH
		JSR	HEXOUT
		LDA	#$00
		JSR	HEXOUT

		JSR	CLOCKSTART	; call clock 

		LDY	#SPACE-STRINGP	
		JSR	WRSTR		; write spacer lines
		LDA	#$1E		; send some space chars to center menu 
		JSR  	TAB		
		LDY	#MENU-STRINGP   ; load menu string
		JSR	WRSTR		; and write it
		LDA	#LOW LANGNAME	; load language name
		STA	PSTRL
		LDA	#HIGH LANGNAME
		STA	PSTRH
		JSR	STROUT		; and write it
		LDA	#SPC
		JSR	COUT
		LDA	#'?'
		JSR	COUT
MLOOP		JSR  	CIN		; main menu loop 
        	AND  	#$DF		; convert the input to uppercase char
        	CMP  	#'M'		; (M)onitor choosen?
		BNE	MNEXT1
STARTMON	JMP  	MONITOR		; yes, start monitor
MNEXT1		CMP	LANGKEY		; compare with language key char
		BNE	MNEXT
		JSR  	CLRSCRN    	; clear screen
		JMP	$B000		; jump to language start
MNEXT		JMP  	MLOOP		; no valid input choosen, try again
SHOWMON		JMP	MONRESET

; Load String Pointer *********************************************************
 
CLRLOADSTR      JSR     CLRSCRN
LOADSTRING	LDA  	#LOW STRINGP 	; load program titles
		STA  	PSTRL
		LDA  	#HIGH STRINGP
		STA  	PSTRH
		RTS

; *****************************************************************************
; command executor
; *****************************************************************************

; print command ***************************************************************

PRINTOUT	LDA	#STDPPORT	; set parallel port as standard out
		JSR	SETSTDOUT
		INY
		LDA  	STRBUF,Y   	; get next input char
		AND  	#$DF		; uppercase chars only
		CMP  	#'D'		; print mem dump?
		BEQ	PRINTDUMP
		JSR	DISASSEM
		JMP	ENDINP
PRINTDUMP	JSR	MEMDUMP		; print memory dump
		BCC	ENDINP 		; normal termination?
		JSR	BEEP		; no, ESC pressed. Beep
ENDINP		JSR	CROUT		; send CR/LF to print last line
		LDA	#STDTTY		; set tty terminal as
		JSR	SETSTDOUT
		JMP	MONINP

; XMODEM load/save command ****************************************************

XMODEM		DEX			; check read/write mode
		BEQ	XMODEML		; read mode?
		LDA	MODE		; no, test if valid address mode
		BEQ	NOTVALID	; not valid, get next input
		JSR	XModemSnd	; call xmodem send
		BEQ	XMODEME		
XMODEML		JSR	XModemRcv	; yes, call xmodem receive
XMODEME		JMP	MONINP		; get next command line

; parallel load/save command **************************************************

PARALLEL	DEX			; check read/write mode
		BEQ	PARALLELL	; read mode?
		LDA	MODE		; no, test if valid address mode
		BEQ	NOTVALID	; not valid, get next input
		JSR	PPORTSAVE	; call save pport ### not implemented yet
		JMP	MONINP
PARALLELL	JSR	PPORTLOAD	; call load pport ### not implemented yet
		JMP	MONINP

; tape load/save command ******************************************************

TAPE		DEX			; check read/write mode
		BEQ	TAPEL		; read mode?
		LDA	MODE		; no, test if valid address mode
		BEQ	NOTVALID	; not valid, get next input
		;JSR	TAPESAVE	; call save tape ### not implemented yet
		JMP	MONINP
TAPEL		;JSR	TAPELOAD	; call load tape ### not implemented yet
		JMP	MONINP	

; load/save command ***********************************************************

LOADSAVE	LDA	MODE		; check address mode
		ASL	A		
		BMI	CHKNEXTCMD	; mode = $C0 (block mode)?
		LDA	#$FF		; no, set end address to $ffff
		STA	NUML
		STA	NUMH
CHKNEXTCMD	INY
		LDA  	STRBUF,Y   	; get next input char
		AND  	#$DF		; uppercase chars only
		CMP  	#'M'		; load/save via xmodem
		BEQ  	XMODEM 		
		CMP  	#'P'		; load/save via parallel port
		BEQ  	PARALLEL 	
		CMP  	#'T'		; load/save via tape
		BEQ  	TAPE
		CMP	#'0'
		BCS	NOTVALID
		DEX			; check load/save mode
		BNE	NOTVALID	; if save mode, just get next input char
		JSR	DISASSEM	; last command was L, so call disassembler
		JMP	MONINP		; we are finnished, get next input line
NOTVALID	LDA	#$00		; no valid command, so restore registers
		TAX
		JMP  	SETMODE2   	; and get next input char

; save command ****************************************************************

SAVE		INX

; load command ****************************************************************

LOAD		INX
		JMP	LOADSAVE

; print command ***************************************************************

PRINT		JMP	PRINTOUT
		
; call a program **************************************************************

RUN		JSR	PRADDRESS
		LDA	#'R'		; print R to signal run mode
		JSR	COUT
		JSR     EXECPROG	
		JMP  	MONINP		; jump back from program call
EXECPROG	JMP  	(ADRL)     	; jump to program address; execute program

; **** Start Of Hex Monitor ***************************************************

; *****************************************************************************

MONITOR 	JSR  	CLRLOADSTR    	; clear screen and load pointer to string table
		LDY  	#MONSTR-STRINGP
        	JSR  	WRSTR		; show monitor title
MONRESET	LDA	#$00		; initialize RIOT...
		STA	PADD		; set PA1...PA7 as input
		STA	PBD		; set all port b pins to 0
		LDA	#$3E		; #$1E -> #$3E
		STA	PBDD		; and PB1...PB5 as output
MONINP		JSR  	CROUT
		LDA  	#PROMPT
		JSR  	COUT		; show monitor prompt
		JSR  	STRIN      	; read input string
		LDY  	#$00       	; reset string index
		TYA			; mode = 0 (none)

MONINIT		TAX
SETADRMODE	STA  	MODE       
SKIPCMDCHR      INY			; increment string index
NEXTCMDCHR	LDA  	STRBUF,Y   	; get next input char
		CMP  	#CR
		BEQ  	ENDCMD 		; end of input line, return to reader routine
		CMP  	#' '
		BEQ  	SKIPCMDCHR 	; ignore spaces
		CMP  	#'.'
		BEQ  	SETBLKMODE 	; block mode
		CMP	#':'
		BEQ	SETADRMODE
		AND  	#$DF		; uppercase chars only
		CMP  	#'L'		; LOAD/LIST command		
		BEQ	LOAD 		; load or list data		
		CMP  	#'S'		; SAVE command		
		BEQ	SAVE		; save data
		CMP	#'P'		; PRINT command
		BEQ	PRINT		; print data		
		CMP  	#'G'		; GO command		
		BEQ  	RUN		; call program
		CMP  	#'M'		; JUNIOR MONITOR command		
		BNE  	NEXTCMD	
		JMP	JCRESET		; execute original junior computer monitor
NEXTCMD		CMP  	#'Q'		; QUIT command		
		BEQ  	MONEND		; exit monitor program
		LDA	MODE		; test if list command pending
		BMI	DUMP		; if mode = $80, dump last line
		STY  	YSAV		; save Y
		JSR  	HEXINPUT   	; read hex number
	 	CPY  	YSAV		; min 1 hex digit entered?
		BEQ  	SKIPCMDCHR     	; no, read next command
		LDA  	MODE
		BNE	SETMODE
STOREADR	LDA  	NUML       	; yes, copy input value to last address
                STA  	ADRL
                LDA  	NUMH
                STA  	ADRH
		LDA	#$00		; line list mode
SETMODE		CMP	#':'		; is it store mode?
		BEQ	STOREDATA	; yes, store data
		ORA	#$80
SETMODE2	STA	MODE
		BNE	NEXTCMDCHR	; branch always
SETBLKMODE	LDA	#$40		; set block list mode
		BNE	SETADRMODE
ENDCMD		LDA	MODE		; test if list command pending
		BMI	DUMP		; yes, dump last line			
CMDEND		JMP	MONINP		; read next command line
MONEND		JMP	MAIN

; store data ******************************************************************

STOREDATA	LDA  	NUML       	; load lower byte of number
                STA  	(ADRL,X)   	; store current store address (X=0)
                INC  	ADRL       	; increment lower store index.
                BNE  	NEXTITEM    	; no overflow
                INC  	ADRH       	; add carry to upper store index
NEXTITEM        JMP  	NEXTCMDCHR    	; get next command string

; call memory dump ************************************************************

DUMP		JSR	MEMDUMP
		BCC	NEXTCMDCHR   	; get next input 	
		BCS	CMDEND		; yes, stop printing memory dump

; print memory dump ***********************************************************

MEMDUMP		STX  	PDBCNT
		JSR	CHKESC		; ESC pressed?
		BCC	PRADR		; no, go on dumping 				
		RTS			; yes, exit leaving carry flag set
PRADR		JSR	PRADDRESS	; print current address

; print current data byte *****************************************************

PRDATA		JSR  	SPCOUT		; print space
                LDA  	(ADRL,X)   	; get data from address (X=0)
                JSR  	HEXOUT     	; print data in hex format
		INC  	PDBCNT     	; increment data counter

; examine next address ********************************************************

ADRNEXT		JSR	CMPADDR		; see if there's more to print
                BCS  	FINISHED?  	; no more data to output 

		JSR	INCADR          ; increment list index
DIVCHK          LDA  	PDBCNT
		CMP  	#$08	
		BNE  	MOD16CHK	; do we need a divider?		
		JSR  	SPCOUT		; yes, print single SPC as block divider
MOD16CHK	LDA  	ADRL       	; if address MOD 16 = 0 start new line
                AND  	#$0F
		BNE  	PRDATA
		JSR  	PRASCII
		JMP  	MEMDUMP		; print next line

; check if line print completed ***********************************************

FINISHED?	LDA  	MODE		; examine last mode
		STX  	MODE       	; set mode 0
		ASL	A		; mode = $D0?
		BPL  	ENDDUMP		; no, get next input
		JSR  	PRASCII		; yes, we are not finished, print ASCII output for last address
ENDDUMP		CLC			; normal exit, so clear carry flag
		RTS

; print a column with ASCII representation of data ****************************

PRASCII         STY  	YSAV       	; store Y
		SEC			; no carry to subtract
		LDA  	#52		; max tabs 
		SBC  	PDBCNT		; calc tab count to print ASCII column 
		SBC  	PDBCNT		; tab = 52-3*printed_data_bytes_count
		SBC  	PDBCNT
		LDY  	PDBCNT
		CPY  	#9		; more than 8 bytes viewed?
		BCS  	NOADJUST	; no
		ADC  	#1		; yes, adjust by one char for block divider
NOADJUST	JSR  	TAB		; print tab spaces
		
		LDY  	#$00	
NEXTASC		LDA  	(ASCL),Y   	; get data from address
		CMP	#$7F
		BCS  	NOASC      	; char >= ASCII 127? yes, print '.'
		CMP  	#' '
		BCS  	ASCOUT		; printable character? 
NOASC		LDA  	#'.'       	; no, print '.'
ASCOUT		JSR  	COUT
		INY  
		CPY  	PDBCNT
		BNE  	NEXTASC
		LDY  	YSAV       	; restore Y
		RTS

; prompt new line with current address ***************************************

;*****************************************************************************

PRADDRESS	JSR  	CROUT
		LDA  	ADRH       
                STA  	ASCH		; store current print address high-order byte
		JSR  	HEXOUT		; print high-order byte of address
                LDA  	ADRL       
		STA  	ASCL		; store current print address low-order byte
                JSR  	HEXOUT		; print low-order byte of address
		LDA  	#ADIV      	; print '-'
                JSR  	COUT
		RTS

; compare if start address ADR is greater end address NUM ********************

;*****************************************************************************

CMPADDR		LDA  	ADRL       	; see if there's more to print
                CMP  	NUML
                LDA  	ADRH
                SBC  	NUMH
		RTS

; jump to original Junior Computer reset vector ******************************

;*****************************************************************************
 
JCRESET		LDA	#$00		
		STA	PBD		; set Port B output pins to 0
		LDA	#$3F		; 
		JMP	RESET		; jump to Junior Computer reset routine + 2


; Try To Detect IO/Language Card *********************************************

; ****************************************************************************

DETECTLNGCARD	LDA	#$00		
		STA	IOBASEL		; set low byte of IO base pointer to $00
		STA	IOBASEH		; set high byte of IO base pointer to $00
		LDA	K4+$20		; walk through all IO spaces
		LDA	K3+$20		; and trigger annunciator address
		LDA	K2+$20		; to initially switch from ROM to RAM
K4?		LDY	#$10		
		LDA	K4+$30		; try switching to ROM at base address $1000.
		JSR	GETMAGIC	; get magic number? 
		BEQ	STOREBASE	; yes, store base address.
K3?		LDY	#$0C		; no, 
		LDA	K3+$30		; try switching to ROM at base address $0C00.
		JSR	GETMAGIC	; get magic number?
		BEQ	STOREBASE	; yes, store base address.
K2?		LDY	#$08		; no, 
		LDA	K2+$30		; try switching to ROM at base address $0800.
		JSR	GETMAGIC	; get magic number?
		BNE	NOCARD		; no, exit
STOREBASE	STY	IOBASEH		; card found, set high byte of base pointer.
		JSR	INIT_IO		; and initialize VIA 1
NOCARD		RTS

GETMAGIC	LDX	#$04
MAGICLOOP	LDA	MAGIC-1,X
		CMP	$DFFB,X
		BNE	NOMAGIC
		DEX	
		BNE	MAGICLOOP
NOMAGIC		TXA
NOSTDPROC	RTS
	
; *****************************************************************************
; String Data Section
; *****************************************************************************

MAGIC		DB	$65,$02,$CA,$FE

PSSTR		TEXT	"PSAYX"				; processor status string

PROC_STD_OUT
PROC_TTY_OUT	DW	SERIALOUT,TTY_HOME,TTY_CLRSCRN,TTY_NORMAL,TTY_INVERSE,TTY_FLASH,NOSTDPROC,NOSTDPROC
PROC_PPORT_OUT	DW	PPORTOUT,NOSTDPROC,NOSTDPROC,NOSTDPROC,NOSTDPROC,NOSTDPROC,NOSTDPROC,NOSTDPROC

STRINGP							; *** string base pointer ***

ESCCLS  	DB     	$1B,$5B,$32,$4A                 ; VT100 clear screen sequence
ESCHOME        	DB     	$1B,$5B,$48,$00                 ; VT100 cursor home sequence
ESCCLL        	DB     	$0D,$1B,$5B,$32,$4B,$00         ; VT100 clear line sequence
ESCGID		DB     	$1B,$5B,$30,$63,$00	        ; VT100 get ID sequence
ESCNORM		DB     	$1B,$5B,$6D,$00	        	; VT100 set normal text mode
ESCINV		DB     	$1B,$5B,$37,$6D,$00	        ; VT100 set inverse text mode
ESCBLNK		DB     	$1B,$5B,$35,$6D,$00	        ; VT100 set blinking text mode

TITLE		TEXT   	"Junior Computer ]["
        	DB     	CR,CR,CR
		TEXT   	" Print Monitor Version "
        	DB     	VERMAIN,$2E,VERSUB,CR
        	TEXT   	" 2020/22 by Joerg Walke"
		DB     	CR,CR,$00
IOCARD		TEXT	" IO/Language-Card at $"
		DB	$00
SPACE    	DB     	CR,CR,CR,CR,$00
MENU		TEXT   	"(M)onitor  "
        	DB     	$00
MONSTR		DB	CR
		TEXT   	"Hex Monitor"
		DB     	CR,$00

DT_NOT_SET	DB	13,13
		TEXT	" Date/Time not set"
		DB	13,0
DATEINPUT	DB	13
		TEXT	" Date: "
		TEXT	"DD.MM.YY"
		DB	8,8,8,8,8,8,8,8,0
TIMEINPUT	DB	13
		TEXT	" Time: "
		TEXT	"HH:MM:SS"
		DB	8,8,8,8,8,8,8,8,0

STRINGP2

DAYS		TEXT	"MON"
		DB	0
		TEXT	"TUE"
		DB	0
		TEXT	"WED"
		DB	0
		TEXT	"THU"
		DB	0
		TEXT	"FRI"
		DB	0
		TEXT	"SAT"
		DB	0
		TEXT	"SUN"
		DB	0

; *****************************************************************************
; START OF DISASSEMBLER
; *****************************************************************************

DISASSEM	LDA	MODE
		ASL	A
		BPL	SHOW1PAGE	; mode <> $C0 (block mode)?
NEXTLINE1	JSR     CHKESC          ; ESC pressed?
		BCS	ENDDISASSEM	; yes, quit disassembling	 		
NEXTOP		JSR	LOADOPCODE	; load current opcode
MORE?		JSR	CMPADDR		; see if there's more to print
                BCC  	NEXTLINE1  	; no more data to output
		RTS	
SHOW1PAGE	LDA	#23		; show 23 disassembled lines
		STA	LINECNT
NEXTLINE2	JSR	LOADOPCODE	; load current opcode
		DEC	LINECNT
		BNE	NEXTLINE2	; more lines to show?
ENDDISASSEM	RTS			; no, jump back to monitor
	
; load next opcode 
		
LOADOPCODE	JSR	PRADDRESS	; print current address
		LDY	#$00
		LDA	(ADRL),Y	; load opcode
		TAY			; store opcode in Y
		AND	#$03	
		CMP	#$03		; is it a unused opcode?
		BNE	DECODE  	; no, decode it
		LDY	#$02		; Y points to unused opcode $02 to print '???'
		TYA			; and we also need it in A
		BNE	DECODEMNEM	; branch always

; decode opcode index into compressed opcode instruction table

DECODE		TYA			; reload opcode into A
		LSR	A		; every fourth column in the opcode table is a (opcode) gap 
		LSR	A		; so we have to adjust the index because these columns are
					; stripped off in our indirect index table 
		STA	TEMP		; store number of bytes to subtract
		TYA			; reload opcode again
		SEC			
		SBC	TEMP		; and subtract value in TEMP from the original opcode. 
DECODEMNEM	STY	OPCODE		; store opcode
		TAY			; Y holds now the actual index to the stripped opcode table
		LDA	OPCODES,Y	; load packed mnemonic_index/instr._bytes from opcode table
		TAY			; and save to Y
		AND	#$03		; the lower two bits are the number of instruction bytes
		STA	IBYTES		; store it in IBYTES var
		TYA			; reload packed index
		LSR	A		; and strip the lower two bits off
		LSR	A		
		TAY			; Y holds now the index to the mnemonics table
		LDA   	MNEMONICSH,Y    ; load first packed byte of mnemonic string
		STA   	ASCH            ; and store it as left byte of mnemonic
		LDA   	MNEMONICSL,Y    ; load second packed byte of mnemonic string
		STA   	ASCL            ; and store it as right byte of mnemonic
		JSR	SHOWHEX		; first print out all instruction bytes as hex numbers
		LDX   	#$03		; we have to unpack three chars
NEXTMCHR        LDA   	#$00		; clear A
		LDY   	#$05            ; shift 5 bits into A
NEXTSHIFT       ASL   	ASCL           	; MSBit of ASCL in C
		ROL   	ASCH		; C in LSBit of ASCH and MSBit of ASCH in C
		ROL   	A		; C in A
		DEY
		BNE   	NEXTSHIFT
		ADC   	#'?'		; add offset to result, to make it an ASCII char
		JSR     COUT		; print one character of mnemonic
		DEX
		BNE     NEXTMCHR	; more chars to decode?
		LDA	#$02		; print two space chars
		JSR	TAB

; decode address mode and print left part of mode string ('#', '(' or 'A')     
                	
		LDX     #$A0            ; default address mode is implied
		LDY	#$FF		
NEXTMASK	INY
		CPY	#$0F		; all masks tested?
		BEQ	ENDMASK		; yes, finish. Address mode is implied
		LDA	ADRMODEM,Y	; load mask
		AND	OPCODE		; mask opcode
		CMP	ADRMODEC,Y	; is it the mask result?
		BNE	NEXTMASK	; no, try next mask
		LDX	ADRMODER,Y	; yes, load the resulting address mode
ENDMASK		STX	ADRMODE		; save address mode
		TXA			; address mode is in A
		AND	#$0F		; A holds left mode string index
		TAY
		LDA	ADRMSTRL,Y      ; load left mode string
		BEQ	PRINTVAL	; is it a NULL char? Then there is nothing to print
		JSR	COUT		; else print character

; print either one or two operand value bytes

PRINTVAL	JSR	INCADR		; increment current address
		LDX	IBYTES		; load number of instruction bytes
		DEX			; more than one IBs?
		BEQ	ENDINC		; no, just finish
		LDA	#'$'		; yes, print operant value
		JSR	COUT		; first print out '$' as hex number indicator
		LDY	#$01		
		DEX			; more than two IBs?
		BEQ	HEX1		; no, just print one byte
		LDA	(ADRL),Y	; load high byte
		JSR	HEXOUT		; and print it as hex number
HEX1		DEY			
		LDA	(ADRL),Y	; load low byte
		LDX	ADRMODE
		CPX	#$A4		; is it a branch opcode?
		BEQ	CALCADR		; yes, calculate branch destination address
		JSR	HEXOUT		; no, print byte as hex number

; print right part of mode string. (',X', ',Y', ',X)', '),Y' or ')' )

		TXA			; load address mode in A
		LSR	A		; upper nibble is index to right address mode string
		LSR	A		; so we have to shift it right by four bits
		LSR	A
		LSR	A		
		TAY
NEXTAMCHR	LDA	ADRMSTRR,Y	; load one char of right mode string
		BEQ     ENDMODE		; if a NULL char then we are finished
		JSR	COUT		; else print char
		INY
		BNE	NEXTAMCHR	; branch always

; finish current instruction

ENDMODE		LDA	IBYTES		; how many address increments left?
		CMP	#$03
		BNE	INCADR		; just one?
		JSR	INCADR		; no, two increments needed

; increment current address

INCADR  	INC  	ADRL    	; increment current address
        	BNE  	ENDINC  	; no carry!
        	INC  	ADRH
ENDINC		RTS			; end of disassembly

; calculate destination address for branch instructions

CALCADR		SEC
		TAY			; transfer A to Y for sign test
		BPL	ADDOFS		; is the branch offset positiv?
		EOR	#$FF		; no, subtract offset from current address
		STA	STOL
		LDA	ADRL		
		SBC	STOL		; subtract branch offset from LSB current address
		TAY			; store low byte of address to Y
		LDA	ADRH
		SBC	#$00		; substract carry from MSB of address
		JMP	PRINTOFFS	
ADDOFS		ADC	ADRL		; add branch offset to LSB of current address
		TAY			; store low byte of address to Y
		LDA	ADRH		
		ADC	#$00		; add carry to MSB of address
PRINTOFFS	JSR	HEXOUT		; print high byte of branch address
		TYA
		JSR	HEXOUT		; print low byte of branch address
		JMP	INCADR		; and increment current address by one

; show instruction bytes as hex values and trailing variable number of space chars 

SHOWHEX		LDY	#$00
NEXTBYTE	JSR	SPCOUT		; print leading space char
		LDA	(ADRL),Y	; load data byte
		JSR	HEXOUT		; and print it
		INY
		CPY	IBYTES		; all data bytes printed?
		BNE	NEXTBYTE	; no, print next byte
		LDA	#$0C		; tab size is 12
		SEC
CALCTAB		SBC	#$03		; reduce tab space by 3 for every data byte
		DEY
		BNE	CALCTAB		; all data bytes considered?
		JSR	TAB
		RTS

; address mode decode tables ***********************************************************

; Mask, Mask Result and Mode tables. If Opcode and Mask = Mask Result then Mode
; each Mode holds two indices (4 bits R | 4 bits L) to the mode string parts

; **************************************************************************************

ADRMODEM	DB	$FF,$FF,$FF,$1F,$1F,$1F,$1F,$1F,$9F,$9F,$1C,$1C,$1C,$1C,$DF ; mask bits

ADRMODEC	DB	$6C,$A2,$BE,$01,$09,$10,$11,$19,$0A,$80,$04,$0C,$14,$1C,$96 ; mask result bits

ADRMODER	DB	$52,$A1,$80,$32,$A1,$A4,$72,$80,$A3,$A1,$A0,$A0,$00,$00,$80 ; packed mode bits

; address mode strings ****************************************************************

ADRMSTRL	DB	$00,$23,$28,$41,$00
		;	 0   #   (   A   0

ADRMSTRR	DB	$2C,$58,$00,$2C,$58,$29,$00,$29,$2C,$59,$00
		;	 ,   X   0   ,   X   )   0   )   ,   Y   0

; mnemonics table *********************************************************************

; three characters packed in two bytes. Each character uses 5 bits, last bit is unused

; *************************************************************************************

; low bytes of table

MNEMONICSL	DB	$48, $CA, $1A, $08, $28, $A4, $AA, $94
     		DB	$CC, $5A, $D8, $C8, $E8, $48, $4A, $54
     		DB	$6E, $A2, $72, $74, $88, $B2, $B4, $26
     		DB	$C8, $F2, $F4, $A2, $26, $44, $72, $74
     		DB	$26, $22, $C4, $44, $62, $44, $62, $1A
     		DB	$26, $54, $68, $C8, $88, $8A, $94, $44
     		DB	$72, $74, $B2, $B4, $32, $44, $68, $84, $00

; high bytes of table

MNEMONICSH	DB	$11, $13, $15, $19, $19, $19, $1A, $1B
     		DB	$1B, $1C, $1C, $1D, $1D, $23, $23, $23
     		DB	$23, $23, $24, $24, $29, $29, $29, $34
     		DB	$53, $53, $53, $5B, $5D, $69, $69, $69
     		DB	$6D, $7C, $84, $8A, $8A, $8B, $8B, $9C
     		DB	$9C, $9D, $9D, $A0, $A1, $A1, $A1, $A5
     		DB	$A5, $A5, $A8, $A8, $AD, $AE, $AE, $AE, $00

; compressed opcode table *************************************************************

; each byte holds a 6 bit index to the mnemonic table and 2 bits instruction byte count
; empty opcode table columns (3,7,B,F) are stripped out

; *************************************************************************************

OPCODES         DB	$29, $8A, $E1, $E1, $8A, $0A, $91, $8A, $09, $E1, $8B, $0B
     		DB	$26, $8A, $E1, $E1, $8A, $0A, $35, $8B, $E1, $E1, $8B, $0B
     		DB	$73, $06, $E1, $1A, $06, $9E, $99, $06, $9D, $1B, $07, $9F
     		DB	$1E, $06, $E1, $E1, $06, $9E, $B1, $07, $E1, $E1, $07, $9F
     		DB	$A5, $5E, $E1, $E1, $5E, $82, $8D, $5E, $81, $6F, $5F, $83
     		DB	$2E, $5E, $E1, $E1, $5E, $82, $3D, $5F, $E1, $E1, $5F, $83
     		DB	$A9, $02, $E1, $E1, $02, $A2, $95, $02, $A1, $6F, $03, $A3
     		DB	$32, $02, $E1, $E1, $02, $A2, $B9, $03, $E1, $E1, $03, $A3
     		DB	$E1, $BE, $E1, $C6, $BE, $C2, $59, $E1, $D5, $C7, $BF, $C3
     		DB	$0E, $BE, $E1, $C6, $BE, $C2, $DD, $BF, $D9, $E1, $BF, $E1
     		DB	$7E, $76, $7A, $7E, $76, $7A, $CD, $76, $C9, $7F, $77, $7B
     		DB	$12, $76, $E1, $7E, $76, $7A, $41, $77, $D1, $7F, $77, $7B
     		DB	$4E, $46, $E1, $4E, $46, $52, $69, $46, $55, $4F, $47, $53
     		DB	$22, $46, $E1, $E1, $46, $52, $39, $47, $E1, $E1, $47, $53
     		DB	$4A, $AE, $E1, $4A, $AE, $62, $65, $AE, $85, $4B, $AF, $63
     		DB	$16, $AE, $E1, $E1, $AE, $62, $B5, $AF, $E1, $E1, $AF, $63


; *****************************************************************************
; START OF REAL TIME CLOCK CODE
; *****************************************************************************

CLOCKSTART	JSR	CHECKDATETIME
		JSR	CROUT
		JMP	PRINTDATETIME
		
CHECKDATETIME	LDA	#$08
		JSR	READCLOCK
		CPY	#$65
		BNE	DATETIMELOST	
		CPX	#$02
		BNE	DATETIMELOST
		RTS

DATETIMELOST	JSR	LOADSTRING
		LDY	#DT_NOT_SET-STRINGP
		JSR	WRSTR

GETDATETIME	JSR	LOADSTRING
		LDA	#':'
		STA	DIVCHAR
		LDY	#TIMEINPUT-STRINGP
		JSR	WRSTR
		LDY	#$23
		JSR	GETDIGIT
		STA	ACC
		JSR	PRINTDIVCHAR
		LDY	#$59
		JSR	GETDIGIT
		STA	XREG
		JSR	PRINTDIVCHAR
		LDY	#$59
		JSR	GETDIGIT
		STA	YREG
		JSR	WRITETIME2

		LDA	#'.'
		STA	DIVCHAR
		LDY	#DATEINPUT-STRINGP
		JSR	WRSTR
		LDY	#$31
		JSR	GETDIGIT
		STA	YREG
		JSR	PRINTDIVCHAR
		LDY	#$12
		JSR	GETDIGIT
		STA	XREG
		JSR	PRINTDIVCHAR
		LDY	#$99
		JSR	GETDIGIT
		STA	ACC
		JSR	WRITEDATE2
		
		RTS

GETDIGIT	INY
		STY	YSAV
GETDIGIT1	JSR	NUMINPUT
		BCC	GETDIGIT1
		TAX
		SBC	#48
		ASL	A
		ASL	A
		ASL	A
		ASL	A	
		CMP	YSAV
		BCS	GETDIGIT1
		STA	TEMP
		TXA
		JSR	COUT
GETDIGIT2	JSR	NUMINPUT
		BCC	GETDIGIT2
		TAX
		SBC	#48
		ORA	TEMP	
		CMP	YSAV
		BCS	GETDIGIT2
		STA	TEMP
		LDA	#'.'
		CMP	DIVCHAR
		BNE	GETDIGITEND
		LDA	TEMP
		BEQ	GETDIGIT2
GETDIGITEND	TXA
		JSR	COUT
		LDA	TEMP
		RTS
		

NUMINPUT	JSR	CIN
		CMP	#'0'
		BCC	NOTNUM
		CMP	#':'
		BCC	ISNUM
NOTNUM		CLC
		RTS
ISNUM		SEC
		RTS		

; *******************************************

PRINTDATETIME	JSR	CROUT
		JSR	SPCOUT
		JSR	PRINTDATE	;PRINTFULLDATE		
		JSR	SPCOUT
		JSR	PRINTTIME
		JSR	CROUT
		RTS

PRINTTIME	LDA	#':'
		STA	DIVCHAR
		JSR	READTIME
		JSR	PRINTDIGIT
		TXA
		JSR	PRINTDIGIT
		TYA
		JSR	HEXOUT
		RTS

PRINTFULLDATE	JSR	READDOW
		ASL	A
		ASL	A
		TAY
		LDA  	#LOW DAYS-4 	
		STA  	PSTRL
		LDA  	#HIGH DAYS-4
		STA  	PSTRH
		JSR	WRSTR
		JSR	SPCOUT
PRINTDATE	LDA	#'.'
		STA	DIVCHAR
		JSR	READDATE
		STA	TEMP
		TYA
		JSR	PRINTDIGIT
		TXA
		JSR	PRINTDIGIT
		LDA	#$20
		JSR	HEXOUT
		LDA	TEMP
		JSR	HEXOUT
		RTS

PRINTDIGIT	JSR	HEXOUT
PRINTDIVCHAR	LDA	DIVCHAR
		JSR	COUT
		RTS

; *****************************************************************************
; LOW LEVEL REAL TIME CLOCK CODE
; *****************************************************************************

; **** Set Day Of Week ********************************************************

; Input: A - Day Of Week 1 (MON) - 7 (SUN)

; *****************************************************************************	

WRITEDOW	STA	ACC
		LDA	#$03
		JSR	SETI2CADR
		LDA	ACC
		JSR	I2C_SEND	; set day of week
		JSR	I2C_STOP
		RTS

; **** Get Day Of Week ********************************************************

; Output: A - Day Of Week 1 (MON) - 7 (SUN)

; *****************************************************************************	

READDOW		LDA	#$03
		JSR	READCLOCK
		TYA
		RTS

; **** Write Time *************************************************************

; Input: A - HOUR 
;	 X - MINUTE 
;	 Y - SECOND

; *****************************************************************************	

WRITETIME	STA	ACC
		STX	XREG
		STY	YREG
WRITETIME2	LDA	#$00		; start at register 0
		JSR	WRITECLOCK	; write time bytes to clock registers
		LDA	#$08		; set address pointer to ram
		JSR	SETI2CADR
		LDA	#$65		; time set mark
		JSR	I2C_SEND
		JSR	I2C_STOP
		RTS

; **** Write Date *************************************************************

; Input: A - YEAR 
;	 X - MONTH 
;	 Y - DAY

; *****************************************************************************	

WRITEDATE	STA	ACC
		STX	XREG
		STY	YREG
WRITEDATE2	LDA	#$04		; start at register 4
		JSR	WRITECLOCK	; write date bytes to clock register
		LDA	#$09		; set address pointer to ram
		JSR	SETI2CADR
		LDA	#$02		; date set mark
		JSR	I2C_SEND
		JSR	I2C_STOP
		RTS

; **** Write Data To Clock ****************************************************

; *****************************************************************************	

WRITECLOCK	JSR	SETI2CADR
		LDA	YREG
		JSR	I2C_SEND	; set second or day
		LDA	XREG
		JSR	I2C_SEND	; set minute or month
		LDA	ACC
		JSR	I2C_SEND	; set hour or year
		JSR	I2C_STOP
		RTS

; **** Read Time **************************************************************

; Output: A - HOUR 
;	  X - MINUTE 
;	  Y - SECOND

; *****************************************************************************	

READTIME	LDA	#$00
		JSR	READCLOCK
		RTS

; **** Read Date **************************************************************

; Output: A - YEAR 
;	  X - MONTH 
;	  Y - DAY

; *****************************************************************************

READDATE	LDA	#$04
		JSR	READCLOCK
		RTS

; **** Read Data From Clock ***************************************************

; *****************************************************************************

READCLOCK	JSR	SETI2CADR
		JSR	I2C_START
		LDA	#I2C_RTC_ADR	; the I2C address		
		JSR	I2C_READ_DEV
		JSR	I2C_RCV
		STA	YREG
		JSR	I2C_ACK
		JSR	I2C_RCV
		STA	XREG
		JSR	I2C_ACK
		JSR	I2C_RCV
		STA	ACC
		JSR	I2C_NACK
		JSR	I2C_STOP
		LDA	ACC
		LDX	XREG
		LDY	YREG
		RTS

; **** Set RTC Address Read/Write Pointer *************************************

; Input: A - Register Address

; *****************************************************************************

SETI2CADR	PHA			; save register address onto stack
		JSR	I2C_START	; send start condition
		LDA	#I2C_RTC_ADR	; the I2C device address	
		JSR	I2C_WRITE_DEV	; send device address and write bit
		PLA			; restore register address
		JSR	I2C_SEND	; send register address
		RTS

; *****************************************************************************
; START OF DATA TRANSFER CODE
; *****************************************************************************

; *****************************************************************************
; Initialize VIA2 Of The IO/Language Card
; *****************************************************************************

INIT_IO		LDY	#DDRB		; initialize data direction of port B
		LDA	#%01011110	; SDA,/CAS_MOT,/CAS_SENSE,CAS_WR,/SPI_LOAD,SPI_CS,/SND_WE,SCL
		STA	(IOBASE),Y		
		LDY	#PORTB
		LDA	#%01001010	; SDA=0,/CAS_MOT=1,CAS_SENSE=0,CAS_WR=0,/SPI_LOAD=1,SPI_CS=0,/SND_WE=1,SCL=0
		STA	(IOBASE),Y	; set SDA as input to pull it high, set SCL as input to pull it high
		LDY	#DDRA		; initialize data direction of port A
		LDA	#$FF		; all pins of port A are outputs
		STA	(IOBASE),Y
		LDY	#PORTA
		LDA	#$00		
		STA	(IOBASE),Y
		RTS

; *****************************************************************************
; START OF VIA1 INPUT/OUTPUT CODE
; *****************************************************************************

; **** Write To VIA1 Register *************************************************

; Input:  Y - Destination Register index (VIA_PORTA,VIA_DDRA...)
; Input:  A - Data to be written into the Register

; *****************************************************************************

WRITE_VIA	STA	(IOBASE),Y
		RTS

; **** Read From VIA1 Register *************************************************

; Input:  Y - Source Register index (VIA_PORTA,VIA_DDRA...)
; Output: A - Read Data from the Register

; *****************************************************************************

READ_VIA	LDA	(IOBASE),Y
		RTS

; *****************************************************************************
; START OF I2C CODE
; *****************************************************************************


; **** Send I2C Start Condition ***********************************************

; *****************************************************************************

I2C_START	LDY	#DDRB
		LDA	#%01011110	; SDA = 1; SCL = 1
		STA	(IOBASE),Y
		LDA	#%11011110	; SDA = 0; SCL = 1
		STA	(IOBASE),Y
		LDA	#%11011111	; SDA = 0; SCL = 0
		STA	(IOBASE),Y
		RTS

; **** Send I2C Stop Condition ************************************************

; *****************************************************************************

I2C_STOP	LDY	#DDRB
		LDA	#%11011111	; SDA = 0; SCL = 0
		STA	(IOBASE),Y
		LDA	#%11011110	; SDA = 0; SCL = 1
		STA	(IOBASE),Y
		LDA	#%01011110	; SDA = 1; SCL = 1
		STA	(IOBASE),Y
		RTS

; **** Send I2C Acknowledged **************************************************

; *****************************************************************************

I2C_ACK		LDY	#DDRB
		LDA	#%11011111	; SDA = 0; SCL = 0
		STA	(IOBASE),Y
		LDA	#%11011110	; SDA = 0; SCL = 1
		STA	(IOBASE),Y
		LDA	#%11011111	; SDA = 0; SCL = 0
		STA	(IOBASE),Y
		RTS

; **** Send I2C Not Acknowledged **********************************************

; *****************************************************************************

I2C_NACK	LDY	#DDRB
		LDA	#%01011111	; SDA = 1; SCL = 0
		STA	(IOBASE),Y
		LDA	#%01011110	; SDA = 1; SCL = 1
		STA	(IOBASE),Y
		LDA	#%01011111	; SDA = 1; SCL = 0
		STA	(IOBASE),Y
		RTS

; **** Read I2C Device ********************************************************

; Input:  A - Device Address
; Output: C - clear not acknowledged, set acknowledged

; *****************************************************************************

I2C_READ_DEV	SEC			; set carry flag
		ROL	A		; shift device address one bit left and rotate C in LSB. LSB = 1 = read
		BNE	I2C_SEND	; and send it

; **** Write I2C Device *******************************************************

; Input:  A - Device Address
; Output: C - clear not acknowledged, set acknowledged

; *****************************************************************************

I2C_WRITE_DEV	ASL	A		; shift device address one bit left. LSB is now 0 = write
					; directly fallthrough to I2C_SEND

; **** Send a Byte to I2C Device **********************************************

; Input:  A - Data Byte
; Output: C - clear not acknowledged, set acknowledged

; *****************************************************************************

I2C_SEND	STA	I2C_DATA
		LDX	#$08		; send 8 bits
		LDY	#DDRB
SENDLOOP	ASL	I2C_DATA	; get next bit into C flag
		BCS	SENDH		; is it a 1 bit?
		LDA	#%11011111	; no, SDA = 0; SCL = 0
		BNE	SETBIT		; branch always
SENDH		LDA	#%01011111	; yes, SDA = 1; SCL = 0
SETBIT		STA	(IOBASE),Y
		AND	#%11111110	; SDA = X; SCL = 1
		STA	(IOBASE),Y
		ORA	#%00000001	; SDA = X, SCL = 0
		STA	(IOBASE),Y
		DEX
		BNE	SENDLOOP

I2C_ACK?	LDY	#DDRB

		LDA	#%01011111	; SDA = 1; SCL = 0
		STA	(IOBASE),Y

		LDA	#%01011110	; SDA = 1; SCL = 1
		STA	(IOBASE),Y
		LDY	#PORTB		
		LDA	(IOBASE),Y	; get SDA
		BPL	ISACK		; SDA = 1 ?
		CLC			; no, not acknowledeged
		BCC	CLKDOWN
ISACK		SEC			; yes, acknowledeged
CLKDOWN		LDY	#DDRB
		LDA	#%01011111	; SCL = 0
		STA	(IOBASE),Y
		RTS

; **** Receive a Byte from I2C Device *****************************************

; Output: A - Data Byte

; *****************************************************************************

I2C_RCV		LDX	#$09		
RCVLOOP		LDY	#DDRB
		LDA	#%01011111	; SDA = 1; SCL = 0
		STA	(IOBASE),Y
		DEX
		BEQ	RCVEND		; all eight bits received?
		LDA	#%01011110	; SDA = 1; SCL = 1
		STA	(IOBASE),Y
		LDY	#PORTB		
		LDA	(IOBASE),Y	; get SDA
		ASL	A		; and shift it into C
		ROL	I2C_DATA	; shift byte buffer one bit left. C goes into LSB
		JMP	RCVLOOP
RCVEND		LDA	I2C_DATA	; load data into A
		RTS

; *****************************************************************************
; START OF XMODEM CODE
; *****************************************************************************
;
; XMODEM/CRC Sender/Receiver for the 6502
;
; By Daryl Rictor Aug 2002
;
; A simple file transfer program to allow transfers between the SBC and a 
; console device utilizing the x-modem/CRC transfer protocol.
;
;**************************************************************************
; This implementation of XMODEM/CRC does NOT conform strictly to the 
; XMODEM protocol standard in that it (1) does not accurately time character
; reception or (2) fall back to the Checksum mode.

; (1) For timing, it uses a crude timing loop to provide approximate
; delays.  These have been calibrated against a 1MHz CPU clock.  I have
; found that CPU clock speed of up to 5MHz also work but may not in
; every case.  Windows HyperTerminal worked quite well at both speeds!
;
; (2) Most modern terminal programs support XMODEM/CRC which can detect a
; wider range of transmission errors so the fallback to the simple checksum
; calculation was not implemented to save space.
;**************************************************************************
;
; Files transferred via XMODEM-CRC will have the load address contained in
; the first two bytes in little-endian format:  
;  FIRST BLOCK
;     offset(0) = lo(load start address),
;     offset(1) = hi(load start address)
;     offset(2) = data byte (0)
;     offset(n) = data byte (n-2)
;
; Subsequent blocks
;     offset(n) = data byte (n)
;
; One note, XMODEM send 128 byte blocks.  If the block of memory that
; you wish to save is smaller than the 128 byte block boundary, then
; the last block will be padded with zeros.  Upon reloading, the
; data will be written back to the original location.  In addition, the
; padded zeros WILL also be written into RAM, which could overwrite other
; data. 
;
;**************************************************************************
;
; Code extensions 2022 by Joerg Walke
;
; Included: CAN command in addition to ESC to cancel sending and receiving data. 
; Included: EOT command to signal end of transmition. 
; Included: address range for received data, to override the start address in
;           the first data block and to prevent overwriting of data by 
;	    trailing zeros.

; XMODEM Receive Routine ******************************************************

XModemRcv       JSR     PrintXStart
		STA	BLKEND		; set flag to false
                LDA     #$01
                STA     BLKNO           ; set block # to 1
                STA	BFLAG           ; set flag to get address from block 1
StartRcv        LDA     #"C"            ; "C" start with CRC mode
                JSR     BOUT            ; send it
                LDA     #$FF    
                STA     RETRYH          ; set loop counter for ~3 sec delay
                LDA     #$00
                STA     CRCL
                STA     CRCH            ; init CRC value        
                JSR     GetByte         ; wait for input
		BCS     GotByte         ; byte received, process it
		JMP     StartRcv
StartBlk        LDA     #$FF             
                STA     RETRYH          ; set loop counter for ~3 sec delay
                JSR     GetByte         ; get first byte of block
                BCC     StartBlk        ; timed out, keep waiting...
GotByte         CMP     #ESC            ; quitting?
                BEQ     GotESC          ; yes
		CMP	#CAN		; cancel?
		BNE     GotByte1	; no
GotESC          JMP     PrintXErr       ; print error and return
GotByte1        CMP     #SOH            ; start of block?
                BEQ     BegBlk          ; yes
                CMP     #EOT            ;
                BNE     BadCRC          ; Not SOH or EOT, so flush buffer & send NAK    
                JMP     RDone           ; EOT - all done!
BegBlk          LDX     #$00
GetBlk          LDA     #$FF            ; 3 sec window to receive characters
                STA     RETRYH          
GetBlk1         JSR     GetByte         ; get next character
                BCC     BadCRC          ; chr rcv error, flush and send NAK
GetBlk2         STA     RBUFF,x         ; good char, save it in the rcv buffer
                INX                     ; inc buffer pointer    
                CPX     #$84            ; <01> <FE> <128 bytes> <CRCH> <CRCL>
                BNE     GetBlk          ; get 132 characters
                LDX     #$00            
                LDA     RBUFF,x         ; get block # from buffer
                CMP     BLKNO           ; compare to expected block #   
                BEQ     GoodBlk1        ; matched!
                jsr     PrintXErr       ; Unexpected block number - abort       
                JMP     Flush           ; mismatched - flush buffer and return
GoodBlk1        EOR     #$FF            ; 1's comp of block #
                INX                     ;
                CMP     RBUFF,x         ; compare with expected 1's comp of block #
                BEQ     GoodBlk2        ; matched!
                JSR     PrintXErr       ; Unexpected block number - abort       
                JMP     Flush           ; mismatched - flush buffer and return      
GoodBlk2        JSR     CalcCRC         ; calc CRC
                LDA     RBUFF,y         ; get hi CRC from buffer
                CMP     CRCH            ; compare to calculated hi CRC
                BNE     BadCRC          ; bad crc, send NAK
                INY                     ;
                LDA     RBUFF,y         ; get lo CRC from buffer
                CMP     CRCL            ; compare to calculated lo CRC
                BEQ     GoodCRC         ; good CRC
BadCRC          JSR     Flush           ; flush the input port
                LDA     #NAK            ;
                JSR     BOUT            ; send NAK to resend block
                JMP     StartBlk        ; start over, get the block again                       
GoodCRC         LDX     #$02            ;
                LDA     BLKNO           ; get the block number
                CMP     #$01            ; 1st block?
                BNE     CopyBlk         ; no, copy all 128 bytes
                LDA     BFLAG           ; is it really block 1, not block 257, 513 etc.
                BEQ     CopyBlk         ; no, copy all 128 bytes
		LDA     MODE		; address mode = 0?
		BEQ	READADR         ; yes, read start address from data stream
                INX
		BNE     READDATA	; branch always
READADR         LDA     RBUFF,x         ; get target address from 1st 2 bytes of blk 1        
		STA     ADRL            ; save lo address
                INX                     
                LDA     RBUFF,x         ; get hi address
                STA     ADRH            ; save it
READDATA        LDA	ADRL
		STA	STOL		; save start address low byte
		LDA	ADRH
		STA	STOH		; save start address high byte
		INX                     ; point to first byte of data
                DEC     BFLAG           ; set the flag so we won't get another address          
CopyBlk         LDY     #$00            ; set offset to zero
CopyBlk3        LDA     BLKEND		; block end flag set?
		BNE     CopyBlk5	; yes, skip reading data
		LDA     RBUFF,x         ; get data byte from buffer
		STA     (STOL),y        ; save to target              
		SEC                      
                LDA     NUML            
                SBC     STOL            ; are we at the last address?
                BNE     CopyBlk5  	; no, inc pointer and continue
                LDA     NUMH           
                SBC     STOH            
                BNE     CopyBlk5   
                INC     BLKEND		; yes, set last byte flag
CopyBlk5	INC     STOL            ; point to next address
                BNE     CopyBlk4        ; did it step over page boundary?
                INC     STOH            ; adjust high address for page crossing
CopyBlk4        INX                     ; point to next data byte
                CPX     #$82            ; is it the last byte
                BNE     CopyBlk3        ; no, get the next one
IncBlk          INC     BLKNO           ; done.  Inc the block #
                LDA     #ACK            ; send ACK
                JSR     BOUT         
                JMP     StartBlk        ; get next block
RDone           LDA     #ACK            ; last block, send ACK and exit.
                JSR     BOUT         
                JSR     Flush           ; get leftover characters, if any
                JMP     PrintXSucc   

; XMODEM Send Routine *********************************************************

XModemSnd       JSR     PrintXStart 
		STA     ERRCNT          ; error counter set to 0
		STA     BLKEND          ; set flag to false
		LDA     #$01            
                STA     BLKNO           ; set block # to 1
Wait4CRC        LDA     #$FF            ; 3 seconds
                STA     RETRYH          
                JSR     GetByte        
                BCC     Wait4CRC        ; wait for something to come in...
                CMP     #"C"            ; is it the "C" to start a CRC xfer?
                BEQ     SetStoAddr      ; yes
                CMP     #ESC            ; is it a cancel? <Esc> Key
                BEQ     DoCancel        ; No, wait for another character
		CMP     #CAN            ; is it a cancel?
                BNE     Wait4CRC        ; No, wait for another character
DoCancel        JMP     PrtAbort        ; Print abort msg and exit
SetStoAddr	LDA     #$01            ; manually load blk number      
                STA     RBUFF           ; into 1st byte
                LDA     #$FE            ; load 1's comp of block #      
                STA     RBUFF+1         ; into 2nd byte
                LDA     ADRL            ; load low byte of start address                
                STA     RBUFF+2         ; into 3rd byte 
                LDA     ADRH            ; load hi byte of start address         
                STA     RBUFF+3         ; into 4th byte
		LDX     #$04            ; preload X to receive buffer
		LDY     #$00            ; init data block offset to 0
                BEQ     LdBuff1         ; jump into buffer load routine 
LdBuffer        LDA     BLKEND          ; was the last block sent?
                BEQ     LdBuff0         ; no, send the next one 
                JMP     SDone           ; yes, we're done
LdBuff0         LDX     #$02            ; init pointers
                LDY     #$00            
                INC     BLKNO           ; inc block counter
                LDA     BLKNO            
                STA     RBUFF           ; save in 1st byte of buffer
                EOR     #$FF             
                STA     RBUFF+1         ; save 1's comp of blkno next
LdBuff1         LDA     (ADRL),y        ; save 128 bytes of data
                STA     RBUFF,x         
LdBuff2         SEC                      
                LDA     NUML            
                SBC     ADRL            ; are we at the last address?
                BNE     LdBuff4         ; no, inc pointer and continue
                LDA     NUMH           
                SBC     ADRH            
                BNE     LdBuff4          
                INC     BLKEND          ; yes, set last byte flag
LdBuff3         INX                     
                CPX     #$82            ; are we at the end of the 128 byte block?
                BEQ     SCalcCRC        ; yes, calc CRC
                LDA     #$00            ; fill rest of 128 bytes with $00
                STA     RBUFF,x         
                BEQ     LdBuff3         ; branch always
LdBuff4         INC     ADRL            ; inc address pointer
                BNE     LdBuff5         
                INC     ADRH            
LdBuff5         INX                     
                CPX     #$82            ; last byte in block?
                BNE     LdBuff1         ; no, get the next
SCalcCRC        JSR     CalcCRC
                LDA     CRCH            ; save hi byte of CRC to buffer
                STA     RBUFF,y         
                INY                     
                LDA     CRCL            ; save lo byte of CRC to buffer
                STA     RBUFF,y
Resend          LDX     #$00            
                LDA     #SOH
                JSR     BOUT            ; send SOH
SendBlk         LDA     RBUFF,x         ; send 132 bytes in buffer to the console
                JSR     BOUT         
                INX                     
                CPX     #$84            ; last byte?
                BNE     SendBlk         ; no, get next
                LDA     #$FF            ; yes, set 3 second delay 
                STA     RETRYH          ; and
                JSR     GetByte         ; wait for ACK/NACK
                BCC     SetError        ; no char received after 3 seconds, resend
                CMP     #ACK            ; char received... is it:
                BEQ     LdBuffer        ; ACK, send next block
                CMP     #NAK             
                BEQ     SetError        ; NAK, inc errors and resend
                CMP     #ESC            
                BEQ     PrtAbort        ; ESC pressed to abort
		CMP	#CAN
		BEQ     PrtAbort	; CANCEL send
					; fall through to error counter
SetError        INC     ERRCNT          ; inc error counter
                LDA     ERRCNT           
                CMP     #$0A            ; are there 10 errors? (Xmodem spec for failure)
                BNE     Resend          ; no, resend block

PrtAbort        JSR     Flush           ; yes, too many errors, flush buffer,
                JMP     PrintXErr       ; print error msg and exit
SDone           ;LDA	#EOT		; send EOT (end of text)
		;JSR	BOUT
		;LDA	#$FF		; and get ACK (no repeated EOT if NAK)
		;JSR	GetByte
		JMP     PrintXSucc   	; All Done..Print msg and exit                  

; Get Byte From Serial Port ***************************************************

GetByte         LDA     #$00            ; wait for chr input and cycle timing loop
                STA     RETRYL          ; set low value of timing loop
LoopGetByte     JSR     CGET            ; get chr from serial port, don't wait 
                BCS     EndGetByte      ; got one, so exit
                DEC     RETRYL          ; no character received, so dec counter
                BNE     LoopGetByte     
                DEC     RETRYH          ; dec hi byte of counter
                BNE     LoopGetByte     ; look for character again
                CLC                     ; if loop times out, CLC, else SEC and return
EndGetByte      RTS                     ; with character in A

; Empty Buffer ****************************************************************

Flush           LDA     #$70            ; flush receive buffer
                STA     RETRYH          ; flush until empty for ~1 sec.
Flush1          JSR     GetByte         ; read the port
                BCS     Flush           ; if char received, wait for another
                RTS 

; Calculate CRC ***************************************************************

CalcCRC		LDA	#$00		; calculate the CRC for the 128 bytes
		STA	CRCL		
		STA	CRCH		
		LDY	#$02		
CalcCRC1	LDA	RBUFF,y
		EOR 	CRCH 		; Quick CRC computation with lookup tables
       		TAX		 	; updates the two bytes at crc & crc+1
       		LDA 	CRCL		; with the byte send in the "A" register
       		EOR 	CRCHI,x
       		STA 	CRCH
      	 	LDA 	CRCLO,x
       		STA 	CRCL
		INY			
		CPY	#$82		; done yet?
		BNE	CalcCRC1	; no, get next
		RTS			; y=82 on exit       

; Print XModem Messages *******************************************************

PrintXStart     LDY     #$00		; load start message
		BEQ     PrintXMsg

PrintXErr       JSR	BEEP
		LDY     #(ERRX-MSGX)	; load error message
		BNE     PrintXMsg

PrintXSucc      LDY     #(SUCCX-MSGX)	; load success message

PrintXMsg	LDA     #LOW MSGX 	; load message string
		STA     PSTRL
		LDA     #HIGH MSGX
		STA     PSTRH
		JSR     WRSTR		; and write it
		LDA     #$00     
		RTS

; *****************************************************************************
; String Data Section
; *****************************************************************************

MSGX            DB      CR
		TEXT	"Begin data transfer. Press <ESC> to abort"
		DB     	CR,$00
ERRX		DB	CR
		TEXT	"Transfer Error"
		DB      CR,$00
SUCCX           DB	EOT,EOT,EOT,CR
		TEXT 	"Transfer Successful"
		DB  	CR,$00


; *****************************************************************************
; Low Byte CRC Lookup Table
; *****************************************************************************

                ORG 	$FA00
CRCLO
 		DB 	$00,$21,$42,$63,$84,$A5,$C6,$E7,$08,$29,$4A,$6B,$8C,$AD,$CE,$EF
 		DB 	$31,$10,$73,$52,$B5,$94,$F7,$D6,$39,$18,$7B,$5A,$BD,$9C,$FF,$DE
 		DB 	$62,$43,$20,$01,$E6,$C7,$A4,$85,$6A,$4B,$28,$09,$EE,$CF,$AC,$8D
 		DB 	$53,$72,$11,$30,$D7,$F6,$95,$B4,$5B,$7A,$19,$38,$DF,$FE,$9D,$BC
 		DB 	$C4,$E5,$86,$A7,$40,$61,$02,$23,$CC,$ED,$8E,$AF,$48,$69,$0A,$2B
 		DB 	$F5,$D4,$B7,$96,$71,$50,$33,$12,$FD,$DC,$BF,$9E,$79,$58,$3B,$1A
 		DB 	$A6,$87,$E4,$C5,$22,$03,$60,$41,$AE,$8F,$EC,$CD,$2A,$0B,$68,$49
 		DB 	$97,$B6,$D5,$F4,$13,$32,$51,$70,$9F,$BE,$DD,$FC,$1B,$3A,$59,$78
 		DB 	$88,$A9,$CA,$EB,$0C,$2D,$4E,$6F,$80,$A1,$C2,$E3,$04,$25,$46,$67
 		DB 	$B9,$98,$FB,$DA,$3D,$1C,$7F,$5E,$B1,$90,$F3,$D2,$35,$14,$77,$56
 		DB 	$EA,$CB,$A8,$89,$6E,$4F,$2C,$0D,$E2,$C3,$A0,$81,$66,$47,$24,$05
 		DB 	$DB,$FA,$99,$B8,$5F,$7E,$1D,$3C,$D3,$F2,$91,$B0,$57,$76,$15,$34
 		DB 	$4C,$6D,$0E,$2F,$C8,$E9,$8A,$AB,$44,$65,$06,$27,$C0,$E1,$82,$A3
 		DB 	$7D,$5C,$3F,$1E,$F9,$D8,$BB,$9A,$75,$54,$37,$16,$F1,$D0,$B3,$92
 		DB 	$2E,$0F,$6C,$4D,$AA,$8B,$E8,$C9,$26,$07,$64,$45,$A2,$83,$E0,$C1
 		DB 	$1F,$3E,$5D,$7C,$9B,$BA,$D9,$F8,$17,$36,$55,$74,$93,$B2,$D1,$F0 

; *****************************************************************************
; Hi Byte CRC Lookup Table
; *****************************************************************************

                ORG 	$FB00
CRCHI
 		DB 	$00,$10,$20,$30,$40,$50,$60,$70,$81,$91,$A1,$B1,$C1,$D1,$E1,$F1
 		DB 	$12,$02,$32,$22,$52,$42,$72,$62,$93,$83,$B3,$A3,$D3,$C3,$F3,$E3
 		DB 	$24,$34,$04,$14,$64,$74,$44,$54,$A5,$B5,$85,$95,$E5,$F5,$C5,$D5
 		DB 	$36,$26,$16,$06,$76,$66,$56,$46,$B7,$A7,$97,$87,$F7,$E7,$D7,$C7
 		DB 	$48,$58,$68,$78,$08,$18,$28,$38,$C9,$D9,$E9,$F9,$89,$99,$A9,$B9
 		DB 	$5A,$4A,$7A,$6A,$1A,$0A,$3A,$2A,$DB,$CB,$FB,$EB,$9B,$8B,$BB,$AB
 		DB 	$6C,$7C,$4C,$5C,$2C,$3C,$0C,$1C,$ED,$FD,$CD,$DD,$AD,$BD,$8D,$9D
 		DB 	$7E,$6E,$5E,$4E,$3E,$2E,$1E,$0E,$FF,$EF,$DF,$CF,$BF,$AF,$9F,$8F
 		DB 	$91,$81,$B1,$A1,$D1,$C1,$F1,$E1,$10,$00,$30,$20,$50,$40,$70,$60
 		DB 	$83,$93,$A3,$B3,$C3,$D3,$E3,$F3,$02,$12,$22,$32,$42,$52,$62,$72
 		DB 	$B5,$A5,$95,$85,$F5,$E5,$D5,$C5,$34,$24,$14,$04,$74,$64,$54,$44
 		DB 	$A7,$B7,$87,$97,$E7,$F7,$C7,$D7,$26,$36,$06,$16,$66,$76,$46,$56
 		DB 	$D9,$C9,$F9,$E9,$99,$89,$B9,$A9,$58,$48,$78,$68,$18,$08,$38,$28
 		DB 	$CB,$DB,$EB,$FB,$8B,$9B,$AB,$BB,$4A,$5A,$6A,$7A,$0A,$1A,$2A,$3A
 		DB 	$FD,$ED,$DD,$CD,$BD,$AD,$9D,$8D,$7C,$6C,$5C,$4C,$3C,$2C,$1C,$0C
 		DB 	$EF,$FF,$CF,$DF,$AF,$BF,$8F,$9F,$6E,$7E,$4E,$5E,$2E,$3E,$0E,$1E
	

		END